@@ -21,6 +21,7 @@ from oslo_config import cfg
 from oslo_policy import opts as policy_opts
 from oslo_policy import policy
 
+from openstack_auth import user as auth_user
 from openstack_auth import utils as auth_utils
 
 LOG = logging.getLogger(__name__)
@@ -57,8 +58,8 @@ def _get_enforcer():
                 LOG.debug("adding enforcer for service: %s" % service)
                 _ENFORCER[service] = enforcer
             else:
-                LOG.warn("policy file for service: %s not found at %s" %
-                         (service, enforcer.policy_path))
+                LOG.warning("policy file for service: %s not found at %s" %
+                            (service, enforcer.policy_path))
     return _ENFORCER
 
 
@@ -74,7 +75,8 @@ def check(actions, request, target=None):
     to policy setting.
 
     :param actions: list of scope and action to do policy checks on,
-        the composition of which is (scope, action)
+        the composition of which is (scope, action). Multiple actions
+        are treated as a logical AND.
 
         * scope: service type managing the policy for action
 
@@ -120,48 +122,103 @@ def check(actions, request, target=None):
     # the service APIs will correct us if we are too permissive.
     if target.get('project_id') is None:
         target['project_id'] = user.project_id
+    if target.get('tenant_id') is None:
+        target['tenant_id'] = target['project_id']
     # same for user_id
     if target.get('user_id') is None:
         target['user_id'] = user.id
-    # same for domain_id
-    if target.get('domain_id') is None:
-        target['domain_id'] = user.domain_id
 
-    credentials = _user_to_credentials(request, user)
+    domain_id_keys = [
+        'domain_id',
+        'project.domain_id',
+        'user.domain_id',
+        'group.domain_id'
+    ]
+    # populates domain id keys with user's current domain id
+    for key in domain_id_keys:
+        if target.get(key) is None:
+            target[key] = user.user_domain_id
+
+    credentials = _user_to_credentials(user)
+    domain_credentials = _domain_to_credentials(request, user)
+    # if there is a domain token use the domain_id instead of the user's domain
+    if domain_credentials:
+        credentials['domain_id'] = domain_credentials.get('domain_id')
 
     enforcer = _get_enforcer()
 
     for action in actions:
         scope, action = action[0], action[1]
         if scope in enforcer:
-            # if any check fails return failure
-            if not enforcer[scope].enforce(action, target, credentials):
-                # to match service implementations, if a rule is not found,
-                # use the default rule for that service policy
-                #
-                # waiting to make the check because the first call to
-                # enforce loads the rules
-                if action not in enforcer[scope].rules:
-                    if not enforcer[scope].enforce('default',
-                                                   target, credentials):
-                        return False
-                else:
+            # this is for handling the v3 policy file and will only be
+            # needed when a domain scoped token is present
+            if scope == 'identity' and domain_credentials:
+                # use domain credentials
+                if not _check_credentials(enforcer[scope],
+                                          action,
+                                          target,
+                                          domain_credentials):
                     return False
+
+            # use project credentials
+            if not _check_credentials(enforcer[scope],
+                                      action, target, credentials):
+                return False
+
         # if no policy for scope, allow action, underlying API will
         # ultimately block the action if not permitted, treat as though
         # allowed
     return True
 
 
-def _user_to_credentials(request, user):
+def _check_credentials(enforcer_scope, action, target, credentials):
+    is_valid = True
+    if not enforcer_scope.enforce(action, target, credentials):
+        # to match service implementations, if a rule is not found,
+        # use the default rule for that service policy
+        #
+        # waiting to make the check because the first call to
+        # enforce loads the rules
+        if action not in enforcer_scope.rules:
+            if not enforcer_scope.enforce('default', target, credentials):
+                is_valid = False
+        else:
+            is_valid = False
+    return is_valid
+
+
+def _user_to_credentials(user):
     if not hasattr(user, "_credentials"):
         roles = [role['name'] for role in user.roles]
         user._credentials = {'user_id': user.id,
                              'token': user.token,
                              'username': user.username,
                              'project_id': user.project_id,
+                             'tenant_id': user.project_id,
                              'project_name': user.project_name,
                              'domain_id': user.user_domain_id,
                              'is_admin': user.is_superuser,
                              'roles': roles}
     return user._credentials
+
+
+def _domain_to_credentials(request, user):
+    if not hasattr(user, "_domain_credentials"):
+        try:
+            domain_auth_ref = request.session.get('domain_token')
+
+            # no domain role or not running on V3
+            if not domain_auth_ref:
+                return None
+            domain_user = auth_user.create_user_from_token(
+                request, auth_user.Token(domain_auth_ref),
+                domain_auth_ref.service_catalog.url_for(interface=None))
+            user._domain_credentials = _user_to_credentials(domain_user)
+
+            # uses the domain_id associated with the domain_user
+            user._domain_credentials['domain_id'] = domain_user.domain_id
+
+        except Exception:
+            LOG.warning("Failed to create user from domain scoped token.")
+            return None
+    return user._domain_credentials
